## clsf-LREG-1: classification with logistic regression
##
## Author: Vlad Popovici
##         Bioinformatics Core Facility
##         Swiss Institute of Bioinformatics
##

require(nnet)
source("kcv.r")
source("perf.r")

## FR.WILCOXON - non-parametric version of t-test.
## Returns the raking of the features.
fr.wilcoxon = function(X, y)
  {
    p = apply(X, 1, function(z) (wilcox.test(z ~ y)$p.value) )

    return (order(p, decreasing=FALSE, na.last=TRUE))
  }

## Find the optimal number of features for LREG, by estimating
## (k-fold CV) its performance with varying number of features.
## The performance criterion is AUC.
tune.lreg = function(X, y, ntop=2:nrow(X), kfold=5, rep=10)
  {
    if (length(ntop) == 1)
      return (ntop[1])                  # nothing to try

    lneg = as.numeric(levels(as.factor(y))[1])   # negative label
    lpos = as.numeric(levels(as.factor(y))[2])   # positive label

    ## IDEA:
    ## Instead of the more logical loop nesting:
    ## for (...all reapeats...) {
    ##   for (...all folds...) {
    ##     for (...all feature numbers...) {
    ##         train/test the classifier
    ##          on current fold, with current no. of features
    ##      }
    ##   }
    ## }
    ##
    ## we change the order of the 2 inner loops so we can rank only once the
    ## features (for the current train/test split) and then we vary the number
    ## of selected features and get the predictions.
    ## In the end, all these results have to be put together...
    ##
    A = matrix(0, nrow=length(ntop), ncol=rep)  # the "performance matrix"
    
    for (r in 1:rep) {                 # do repeated CV  
      cv = kcv.splits(y, k=kfold, stratified=TRUE)

      P = matrix(0, nrow=ncol(X), ncol=length(ntop))
      U = rep(FALSE, ncol(X))
      
      for (k in 1:kfold) {
        itr = cv$cv.train[[k]]          # index of elements for training
        ivd = cv$cv.valid[[k]]          # index of elements for validation
        Xtr = X[,itr]; ytr = y[itr]
        Xvd = X[,ivd]; yvd = y[ivd]
        
        # Rank the features
        idx =  fr.wilcoxon(Xtr, ytr)

        U[ivd] = TRUE
        
        kf = 1
        for (nf in ntop) {                   # for each number of feature, estimate the AUC
          Xtr1 = Xtr[idx[1:nf],]
          Xvd1 = Xvd[idx[1:nf],]

          m = multinom(y ~ ., data.frame(y=ytr, t(Xtr1)))
          
          P[ivd,kf] = predict(m, data.frame(t(Xvd1)), type="probs")
          kf = kf + 1
        }
      }
      A[,r] = apply(P, 2, function(p) (perf.auc(p[U], y[U], lneg=lneg, lpos=lpos)))
    }
    # Now A contains the AUC for each feature number and validation fold.
    # Average the AUC for each no. of features and get the optimal one:
    A = apply(A, 1, mean, na.rm=TRUE)

    return (ntop[which.max(A)])
  }
  

## CVESTIM.LREG - estimate the performance of the LREG using 
## repeated 5-fold CV.
cvestim.lreg = function(X, y, ntop=2:nrow(X), kfold=5, rep=10, kcv.file="")
{
  r = matrix(0, 2, 2)                 # confusion matrix
  p.error = vector(mode="numeric", rep)
  p.spec  = vector(mode="numeric", rep)
  p.sens  = vector(mode="numeric", rep)
  p.auc   = vector(mode="numeric", rep)
  p.rmse  = vector(mode="numeric", rep)
  p.mcc   = vector(mode="numeric", rep)
  
  d = nrow(X)          # total number of features
  lneg = as.numeric(levels(as.factor(y))[1])   # negative label
  lpos = as.numeric(levels(as.factor(y))[2])   # positive label

  for (n in 1:rep) {
    cv = kcv.splits(y, k=kfold, stratified=TRUE)
    r.cv = matrix(0, 2, 2)
    used = rep(FALSE, n)
    prob = rep(0, n)

    cat(paste("Repeat", n, "of", rep, "\n"))
    for (k in 1:kfold) {
      itr = cv$cv.train[[k]]          # index of elements for training
      ivd = cv$cv.valid[[k]]          # index of elements for validation
      Xtr = X[,itr]; ytr = y[itr]
      Xvd = X[,ivd]; yvd = y[ivd]

      cat(paste("    Fold:", k, "\n"))
      print("        Tuning...")
      nf = tune.lreg(Xtr, ytr, ntop, kfold=5, rep=5)
      cat(paste("        Best model: top=", nf, "features.\n"))

      idx =  fr.wilcoxon(Xtr, ytr)
        
      Xtr = Xtr[idx[1:nf],]
      Xvd = Xvd[idx[1:nf],]

      used[ivd] = TRUE

      m = multinom(y ~ ., data.frame(y=ytr, t(Xtr)))
      pred = as.numeric(as.character(predict(m, data.frame(t(Xvd)), type="class")))
      prob[ivd] = predict(m, data.frame(t(Xvd)), type="probs")

      r.cv = r.cv + perf.matrix(yvd, pred, lneg=0, lpos=1)
    }
    r = r + r.cv
    ## after each repeat, estimate the current performance
    r.cv = r.cv / kfold
    p.error[n] = perf.error(r.cv)
    p.sens[n]  = perf.sensitivity(r.cv)
    p.spec[n]  = perf.specificity(r.cv)
    p.mcc[n]   = perf.mcc(r.cv)
    
    prb = prob[used]; prb = prb[!is.na(prb)]
    yt  = y[used]; yt = yt[!is.na(prob[used])]
    p.auc[n]   = perf.auc(prb, yt, lneg=0, lpos=1)
    p.rmse[n]  = perf.rmse(prb, yt)
    
    cat(paste("--> Error =", round(100*p.error[n],4),
              "Sensitivity =", round(100*p.sens[n],4),
              "Specificity =", round(100*p.spec[n],4),
              "AUC =", round(100*p.auc[n],4), "\n"))
  }
  r = r / (rep*kfold)
  cat("############################## CV overall performance ##############################:\n")
  cat(paste("  Error = ", round(100*mean(p.error, na.rm=TRUE),4),
    " (", round(100*sd(p.error, na.rm=TRUE),4), ")  ",
    "Sensitivity = ", round(100*mean(p.sens, na.rm=TRUE),4),
    " (", round(100*sd(p.sens, na.rm=TRUE),4), ")  ",
    "Specificity = ", round(100*mean(p.spec, na.rm=TRUE),4),
    " (", round(100*sd(p.spec, na.rm=TRUE),4), ")\n",
    "AUC = ", round(100*mean(p.auc, na.rm=TRUE),4),
    " (", round(100*sd(p.auc, na.rm=TRUE),4), ")\n", sep=""))
  cat("####################################################################################\n")

  return (list(cmat=r, err=p.error, sens=p.sens, spec=p.spec, auc=p.auc, rmse=p.rmse, mcc=p.mcc))
}

## APPLY.LREG - trains the LREG classifier on a training data set and returns the
## posterior probabilities on the test set.
apply.lreg = function(Xtr, ytr, Xts, ntop=2:nrow(X), name="")
{
    print("        Tuning...")
    nf = tune.lreg(Xtr, ytr, ntop, kfold=5, rep=5)
    cat(paste("        Best model: top=", nf, "features.\n"))

    idx =  fr.wilcoxon(Xtr, ytr)

    Xtr = Xtr[idx[1:nf],]
    Xts = Xts[idx[1:nf],]

    m = multinom(y ~ ., data.frame(y=ytr, t(Xtr)))

    pred = as.numeric(as.character(predict(m, data.frame(t(Xts)), type="class")))
    prob = predict(m, data.frame(t(Xts)), type="probs")

  return (prob) # return the posterior probabilities
}
