## DLDA.R: Diagonal Linear Discriminant Analysis
## 
## Author: Vlad Popovici
##         Bioinformatics Core Facility
##         Swiss Institute of Bioinformatics
###############################################################################

##
## DLDA: builds a DLDA model based on the data matrix X (samples by columns)
## and label vector y.
##
dlda = function(X, y)
{
	n  = ncol(X)                        # number of samples
	p  = nrow(X)                        # number of fetures
	nk = rep(0, max(y)-min(y)+1)
	K  = length(nk)
	m  = matrix(0, p, K)
	v  = matrix(0, p, K)
	
	for (k in 1:K) {                    # for each class
		idx = (y == k+min(y)-1)
		nk[k] = sum(idx)
		m[,k] = apply(X[,idx], 1, mean, na.rm=TRUE)
		v[,k] = apply(X[,idx], 1, var, na.rm=TRUE)
	}
	
	vp = apply(v, 1, function(z) (sum((nk-1)*z)/(n-sum(nk!=0))) )
	
	model = list(K=K, m=m, vp=vp, classes=min(y):max(y))
	
	return (model)
}

##
## DLDA.PREDICT predicts the labels for a new data set. It also returns the
## corresponding scores (between 0 and 1).
##
dlda.predict = function(X, model)
{
	K = length(model$classes)
	d = matrix(0, ncol(X), K)
	
	for (k in 1:K) {
		d[,k] = apply(X, 2, function(z) (sum((z-model$m[,k])^2/model$vp)) )
	}
	pred = apply(d, 1, function(z) (model$classes[which.min(z)]) )
	
	# normalize distances, to avoid over- or under-flow
	dmin = min(d)
	dmax = max(d)
	if (dmin != dmax) {
		d = (d - dmin) / (dmax - dmin)
	}
	prob = exp(-d)
	prob = prob / apply(prob, 1, sum)
	
	return (list(pred=pred, prob=prob))
}

##
## DLDA.TRAIN.AND.PREDICT: builds a DLDA model based on the data matrix Xtrain
## (samples by columns) and label vector y and predicts the labels of the samples
## in Xtest.
##
dlda.train.and.predict = function(Xtrain, y, Xtest)
{
	n  = ncol(Xtrain)                   # number of samples
	p  = nrow(Xtrain)                   # number of fetures
	nk = rep(0, max(y)-min(y)+1)
	K  = length(nk)
	m  = matrix(0, p, K)
	v  = matrix(0, p, K)
	
	for (k in 1:K) {                    # for each class
		idx = (y == k+min(y)-1)
		nk[k] = sum(idx)
		m[,k] = apply(Xtrain[,idx], 1, mean, na.rm=TRUE)
		v[,k] = apply(Xtrain[,idx], 1, var, na.rm=TRUE)
	}
	
	vp = apply(v, 1, function(z) (sum((nk-1)*z)/(n-sum(nk!=0))) )
	
	d = matrix(0, ncol(Xtest), K)       # distances from samples to classes
	
	for (k in 1:K) {
		d[,k] = apply(Xtest, 2, function(z) (sum((z-m[,k])^2/vp)) )
	}
	
	pred = apply(d, 1, function(z) ((min(y):max(y))[which.min(z)]) )
	# normalize distances, to avoid over- or under-flow
	dmin = min(d)
	dmax = max(d)
	if (dmin != dmax) {
		d = (d - dmin) / (dmax - dmin)
	}
	prob = exp(-d)
	prob = prob / apply(prob, 1, sum)
	
	return (list(pred=pred, prob=prob))
}

