## FRANK.R - feature ranking functions based on univariate analyses.
##
## Author: Vlad Popovici
##         Bioinformatics Core Facility
##         Swiss Institute of Bioinformatics
################################################################################

require(limma)

## All these functions have at least 2 parameters:
##    X  - a data matrix, with samples by rows, features by columns
##    y  - a vector of labels
##
## Also, all these functions return a permutation of the feature indexes, such that
## the features are ranked according to their importance (e.g. from smallest p-value
## to the largest one), the NAs being left at the end.


## FR.TTEST - ranking the features based on two sample unequal variances t-test.
fr.ttest = function(X, y)
  {
    p = apply(X, 1, function(z) (t.test(z ~ y)$p.value) )

    return (order(p, decreasing=FALSE, na.last=TRUE))
  }


## FR.WILCOXON - non-parametric version of t-test.
fr.wilcoxon = function(X, y)
  {
    p = apply(X, 1, function(z) (wilcox.test(z ~ y)$p.value) )

    return (order(p, decreasing=FALSE, na.last=TRUE))
  }

## FR.BWSS - use the ratio of between to within sum of squares
fr.bwss = function(X, y)
  {
    bw = bwss(X, y)

    return (order(bw, decreasing=TRUE, na.last=TRUE))
  }

## FR.MODTTEST use moderated t-test (from LIMMA)
fr.modttest = function(X, y)
  {
    design = cbind(mean=0, diff=y)
    fit = lmFit(X, design)
    fit =eBayes(fit)
  }



## Function for computing the ratio of between-groups to
## within-groups sum of squares.
##
## TAKEN FROM THE SMA PACKAGE (BIOCONDUCTOR).
# \references{S. Dudoit, J. Fridlyand, and T. P. Speed. Comparison of
# Discrimination Methods for the Classification of Tumors Using Gene
# Expression Data. June 2000. (Statistics, UC Berkeley,  Tech Report \#
# 576). } 
# 
# \author{
#   Sandrine Dudoit, \email{sandrine@stat.berkeley.edu}  \cr
#   Jane Fridlyand, \email{janef@stat.berkeley.edu}
# }

bwss = function(x, cl)
 {
   ## Compute BSS/WSS for each row of a matrix which may have NA
   ## Columns have labels cl=consecutive integers
 
   K <- max(cl) - min(cl) + 1
   tvar <- apply(x, 1, var.na)
   tn <- apply(!is.na(x), 1, sum)
   wvar <- matrix(0, nrow(x), K)
   wn <- matrix(0, nrow(x), K)
   
   for(i in (1:K)) {
     if(sum(cl == (i + min(cl) - 1)) == 1) {
       wvar[, i] <- 0
       wn[, i] <- 1
     }
     if(sum(cl == (i + min(cl) - 1)) > 1) {
       wvar[, i] <- apply(x[, cl == (i + min(cl) - 1)], 1, var.na)
       wn[, i] <- apply(!is.na(x[, cl == (i + min(cl) - 1)]), 1, sum)
     }
   }
   
   WSS <- apply(wvar * (wn - 1), 1, sum.na)
   TSS <- tvar * (tn - 1)
   BSS<- TSS - WSS
   BW <- BSS/WSS
   
   return (BW)
 }
 
## Other functions taken from the same SMA package.
mean.na = function(x,...)
{
  mean(x[!(is.na(x) | is.infinite(x))])
}
 
var.na <- function(x)
{
  res <- NA
  tmp <- !(is.na(x) | is.infinite(x))
  if(sum(tmp) > 1)
    res <- var(x[tmp])
  res
}

cor.na <- function(x)
{
  cor(x, use="pairwise.complete.obs")
}

sum.na <- function(x,...)
{
  res <- NA
  tmp <- !(is.na(x) | is.infinite(x))
  if(sum(tmp) > 0)
    res <- sum(x[tmp])
  res
}


length.na <- function(x, ...)
{
  tmp <- !(is.na(x) | is.infinite(x))
  length(x[tmp],...)
}

log.na <- function(x, ...)
{
  log(ifelse(x > 0, x, NA), ...)
}
